/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.model.framestore;

import java.lang.reflect.*;
import java.util.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.util.*;

/**
 * All queries go to all frame stores.  Writes go to the primary (delegate) frame store.
 * 
 * @author Ray Fergerson <fergerson@smi.stanford.edu>
 */
public class _MergingFrameStoreHandler extends AbstractFrameStoreInvocationHandler {
    private Collection secondaryFrameStores = new ArrayList();

    public static FrameStore newInstance() {
        return newInstance(_MergingFrameStoreHandler.class);
    }

    public void addSecondaryFrameStore(FrameStore fs) {
        secondaryFrameStores.add(fs);
    }

    public void removeSecondaryFrameStore(FrameStore fs) {
        secondaryFrameStores.remove(fs);
    }

    protected Object handleInvoke(Method method, Object[] arguments) {
        Object returnValue;
        if (isQuery(method)) {
            returnValue = handleQuery(method, arguments);
        } else {
            returnValue = handleModification(method, arguments);
        }
        return returnValue;
    }

    protected Object handleQuery(Method method, Object[] arguments) {
        Object returnValue = invoke(method, arguments);
        Iterator i = secondaryFrameStores.iterator();
        while (i.hasNext()) {
            FrameStore frameStore = (FrameStore) i.next();
            Object secondaryReturnValue = invoke(method, arguments, frameStore);
            returnValue = merge(returnValue, secondaryReturnValue);
        }
        return returnValue;
    }

    protected Object merge(Object primary, Object secondary) {
        Object returnValue = null;
        if (primary == null) {
            returnValue = secondary;
        } else if (secondary == null) {
            returnValue = primary;
        } else if (primary instanceof String) {
            returnValue = mergeStrings((String) primary, (String) secondary);
        } else if (primary instanceof Frame) {
            returnValue = mergeFrames((Frame) primary, (Frame) secondary);
        } else if (primary instanceof Integer) {
            returnValue = mergeIntegers((Integer) primary, (Integer) secondary);
        } else if (primary instanceof List) {
            returnValue = mergeLists((List) primary, (List) secondary);
        } else if (primary instanceof Set) {
            returnValue = mergeSets((Set) primary, (Set) secondary);
        } else if (primary instanceof Collection) {
            returnValue = mergeCollections((Collection) primary, (Collection) secondary);
        } else {
            throw new UnsupportedOperationException(primary.toString());
        }
        return returnValue;
    }

    private String mergeStrings(String s1, String s2) {
        Log.warning("Unable to merge", this, "mergeStrings", s1, s2);
        return s1;
    }

    private Frame mergeFrames(Frame f1, Frame f2) {
        Log.warning("Unable to merge", this, "mergeFrames", f1, f2);
        return f1;
    }

    private Integer mergeIntegers(Integer i1, Integer i2) {
        return new Integer(i1.intValue() + i2.intValue());
    }

    private List mergeLists(List list1, List list2) {
        Log.enter(this, "mergeLists", list1 + ", " + list2);
        List merged;
        if (list1.isEmpty()) {
            merged = list2;
        } else if (list2.isEmpty()) {
            merged = list1;
        } else {
            merged = new ArrayList();
            merged.addAll(list1);
            merged.addAll(list2);
        }
        return merged;
    }

    private Set mergeSets(Set set1, Set set2) {
        Log.enter(this, "mergeSets", set1 + ", " + set2);
        Set merged;
        if (set1.isEmpty()) {
            merged = set2;
        } else if (set2.isEmpty()) {
            merged = set1;
        } else {
            merged = new LinkedHashSet();
            merged.addAll(set1);
            merged.addAll(set2);
        }
        return merged;
    }

    private Collection mergeCollections(Collection collection1, Collection collection2) {
        Log.enter(this, "mergeCollections", collection1 + ", " + collection2);
        Collection merged;
        if (collection1.isEmpty()) {
            merged = collection2;
        } else if (collection2.isEmpty()) {
            merged = collection1;
        } else {
            merged = new ArrayList();
            merged.addAll(collection1);
            merged.addAll(collection2);
        }
        return merged;
    }

    protected Object handleModification(Method method, Object[] arguments) {
        return invoke(method, arguments);
    }

}