/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.ui;

import java.awt.event.*;
import java.net.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;

import edu.stanford.smi.protege.action.*;
import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.plugin.*;
import edu.stanford.smi.protege.resource.*;
import edu.stanford.smi.protege.util.*;

/**
 * The main menu bar for the application.
 *
 * @author Ray Fergerson
 */
public class ProjectMenuBar extends JMenuBar {

    private ButtonGroup _group = new ButtonGroup();
    private Map uris = new HashMap();
    private static final int MAX_LENGTH = 50;

    public ProjectMenuBar() {
        createProjectMenu();
        createEditMenu();
        createWindowMenu();
        createHelpMenu();
    }

    private void createEditMenu() {
        JMenu menu = ComponentFactory.createMenu("Edit");
        createItem(menu, new UndoAction());
        createItem(menu, new RedoAction());
        createItem(menu, new ShowCommandHistoryAction());
        add(menu);
    }

    private void createHelpMenu() {
        JMenu menu = ComponentFactory.createMenu("Help");
        add(menu);

        ComponentFactory.addMenuItemNoIcon(
            menu,
            new DisplayHtml("Getting Started", ApplicationProperties.getGettingStartedURLString()));
        ComponentFactory.addMenuItemNoIcon(
            menu,
            new DisplayHtml("Frequently Asked Questions", ApplicationProperties.getFAQURLString()));
        ComponentFactory.addMenuItemNoIcon(
            menu,
            new DisplayHtml("User's Guide", ApplicationProperties.getUsersGuideURLString()));
        ComponentFactory.addMenuItemNoIcon(
            menu,
            new DisplayHtml("Ontology Development 101", ApplicationProperties.getOntology101URLString()));

        addPluginDocSubmenu(menu);

        menu.addSeparator();
        ComponentFactory.addMenuItemNoIcon(menu, new ShowIconDialog());
        menu.addSeparator();

        ComponentFactory.addMenuItemNoIcon(menu, new ShowEncodingAndLocales());
        ComponentFactory.addMenuItemNoIcon(menu, new CollectGarbageAction());

        menu.addSeparator();
        ComponentFactory.addMenuItemNoIcon(menu, new ShowAboutBox());
        ComponentFactory.addMenuItemNoIcon(menu, new ShowAboutPluginsBox());

    }

    private void addPluginDocSubmenu(JMenu helpMenu) {
        JMenu menu = ComponentFactory.createMenu("Plug-ins");
        ComponentFactory.addSubmenu(helpMenu, menu);
        Iterator i = PluginUtilities.getPluginComponentNames().iterator();
        while (i.hasNext()) {
            String name = (String) i.next();
            URL url = PluginUtilities.getPluginComponentDocURL(name);
            if (url != null) {
                ComponentFactory.addMenuItemNoIcon(menu, new DisplayHtml(name, url.toString()));
            }
        }
        if (menu.getItemCount() == 0) {
            menu.setEnabled(false);
        }
    }

    private void createItem(JMenu menu, Action action) {
        ComponentFactory.addMenuItem(menu, action);
    }

    private JMenuItem createItemX(UIManager.LookAndFeelInfo lf) {
        Action action = new LookAndFeelAction(lf);
        JRadioButtonMenuItem item = ComponentFactory.createRadioButtonMenuItem(action);
        _group.add(item);
        if (lf.getClassName().equals(UIManager.getLookAndFeel().getClass().getName())) {
            _group.setSelected(item.getModel(), true);
        }
        return item;
    }

    private JMenu createLAFMenu() {
        JMenu menu = ComponentFactory.createMenu("Look And Feel");

        UIManager.LookAndFeelInfo[] lookAndFeels = UIManager.getInstalledLookAndFeels();
        for (int i = 0; i < lookAndFeels.length; ++i) {
            menu.add(createItemX(lookAndFeels[i]));
        }
        return menu;
    }

    private void createProjectMenu() {
        JMenu menu = ComponentFactory.createMenu(LocalizedText.getProjectMenu());
        loadProjectMenu(menu);
        add(menu);
    }

    private void loadProjectMenu(JMenu menu) {
        menu.removeAll();
        createItem(menu, new CreateProject());
        createItem(menu, new OpenProject());
        createItem(menu, new OpenRemoteProject());
        loadOpenRecent(menu);
        createItem(menu, new SaveProject());
        createItem(menu, new SaveAsProject());
        ComponentFactory.addSubmenu(menu, createArchiveSubmenu());
        createItem(menu, new CloseProject());
        menu.addSeparator();
        createItem(menu, new BuildProject());
        ComponentFactory.addSubmenu(menu, createImportSubmenu());
        ComponentFactory.addSubmenu(menu, createExportSubmenu());
        createItem(menu, new ChangeProjectStorageFormat());
        menu.addSeparator();
        createItem(menu, new IncludeProject());
        createItem(menu, new ChangeIncludedProjects());
        createItem(menu, new ShowIncludedProjects());
        createItem(menu, new MergeIncludedProjects());
        menu.addSeparator();
        createItem(menu, new ConfigureProject());
        createItem(menu, new ShowMetrics());
        createItem(menu, new GenerateHtml());
        menu.addSeparator();
        createItem(menu, new ExitApplication());
    }

    private JMenu createArchiveSubmenu() {
        JMenu menu = ComponentFactory.createMenu(LocalizedText.getArchiveProject());
        menu.setEnabled(ProjectManager.getProjectManager().getCurrentProject() != null);
        createItem(menu, new ArchiveProject());
        createItem(menu, new RevertProject());
        createItem(menu, new ConfigureArchive());
        return menu;
    }

    private JMenu createImportSubmenu() {
        JMenu menu = ComponentFactory.createMenu(LocalizedText.getImportProject());
        Collection classNames = PluginUtilities.getAvailableImportPluginClassNames();
        Iterator i = getSortedPlugins(classNames).iterator();
        while (i.hasNext()) {
            ImportPlugin plugin = (ImportPlugin) i.next();
            createItem(menu, new ImportPluginAction(plugin));
        }
        if (menu.getItemCount() == 0) {
            menu.setEnabled(false);
        }
        return menu;
    }

    private JMenu createExportSubmenu() {
        JMenu menu = ComponentFactory.createMenu(LocalizedText.getProjectMenuExport());
        menu.setEnabled(ProjectManager.getProjectManager().getCurrentProject() != null);
        Collection classNames = PluginUtilities.getAvailableExportPluginClassNames();
        Iterator i = getSortedPlugins(classNames).iterator();
        while (i.hasNext()) {
            ExportPlugin plugin = (ExportPlugin) i.next();
            createItem(menu, new ExportPluginAction(plugin));
        }
        if (menu.getItemCount() == 0) {
            menu.setEnabled(false);
        }
        return menu;
    }

    private List getSortedPlugins(Collection classNames) {
        List plugins = getPlugins(classNames);
        Collections.sort(plugins, new PluginComparator());
        return plugins;
    }

    private List getPlugins(Collection classNames) {
        List plugins = new ArrayList();
        Iterator i = classNames.iterator();
        while (i.hasNext()) {
            String name = (String) i.next();
            Plugin plugin = (Plugin) SystemUtilities.newInstance(name);
            if (plugin != null) {
                plugins.add(plugin);
            }
        }
        return plugins;
    }

    private boolean exists(URI uri) {
        return true;
    }

    private void loadOpenRecent(JMenu menu) {
        JMenu subMenu = ComponentFactory.createMenu(LocalizedText.getOpenRecentProject());
        uris.clear();

        List projectURIList = ApplicationProperties.getMRUProjectList();
        subMenu.setEnabled(!projectURIList.isEmpty());
        for (int i = 0; i < projectURIList.size(); i++) {
            URI uri = (URI) projectURIList.get(i);
            boolean exists = exists(uri);
            if (exists) {
                String text = uri.toString();

                // If the file path is too long to display, reduce it.
                if (text.length() > MAX_LENGTH) {
                    ArrayList list = buildList(text);
                    list = reduce(list);
                    text = rebuildPath(list);
                }

                // Map display text to URI
                uris.put(text, uri);

                JMenuItem menuItem = ComponentFactory.createMenuItem(text);
                menuItem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        JMenuItem source = (JMenuItem) (e.getSource());
                        String text = source.getText();
                        boolean closed = ProjectManager.getProjectManager().closeProjectRequest();
                        if (closed) {
                            URI uri = (URI) uris.get(text);
                            ProjectManager.getProjectManager().loadProject(uri);
                            ApplicationProperties.addProjectToMRUList(uri);
                        }
                    }
                });
                subMenu.add(menuItem);
            }
        }

        ComponentFactory.addSubmenu(menu, subMenu);
    }

    private ArrayList buildList(String path) {
        ArrayList list = new ArrayList();
        StringTokenizer st = new StringTokenizer(path, "/");
        for (int i = 0; st.hasMoreElements(); i++) {
            list.add(i, st.nextElement());
        }
        return list;
    }

    private ArrayList reduce(ArrayList list) {
        // sanity checks
        if (list == null) {
            return null;
        }
        if (list.size() == 0) {
            return list;
        }
        if (getLength(list) <= MAX_LENGTH) {
            return list;
        }

        // size is 1
        if (list.size() == 1) {
            String fname = (String) list.get(0);
            fname = fname.substring(0, (MAX_LENGTH - 3)) + "...";
            list.set(0, fname);
        }
        // size is 2
        else if (list.size() == 2) {
            if (list.get(0) == "...") {
                list.remove(0);
            } else {
                list.set(0, "...");
            }
        }
        // size is 3 or more
        else {
            if (!list.contains("...")) {
                int mid = list.size() / 2;
                list.set(mid, "...");
            } else {
                int idxDot = list.indexOf("...");
                int before = idxDot;
                int after = (list.size() - 2) - idxDot;

                if (before > after) {
                    list.remove(idxDot - 1);
                } else {
                    // since we always remove "after" when before == after,
                    // we are sure that before is always equal to or one more
                    // than before.
                    list.remove(idxDot + 1);
                }
            }
        }

        if (getLength(list) <= MAX_LENGTH) {
            // list is short enough
            return list;
        } else {
            // list is too long - recurse
            return reduce(list);
        }
    }

    private int getLength(ArrayList list) {
        String path = "";
        String sep = "/";
        for (int i = 0; i < list.size(); i++) {
            path = path + list.get(i) + sep;
        }
        return path.length() - 1;
    }

    private String rebuildPath(ArrayList list) {
        String path = "";
        String sep = "/";
        for (int i = 0; i < list.size(); i++) {
            path = path + list.get(i) + sep;
        }
        path = path.substring(0, (path.length() - 1));
        return path;
    }

    private void createWindowMenu() {
        final JMenu menu = ComponentFactory.createMenu("Window");
        menu.addMenuListener(new MenuListener() {
            public void menuSelected(MenuEvent event) {
                loadWindowMenu(menu);
            }

            public void menuCanceled(MenuEvent event) {
            }

            public void menuDeselected(MenuEvent event) {
            }
        });
        add(menu);
    }

    private void loadWindowMenu(JMenu windowMenu) {
        windowMenu.removeAll();
        createItem(windowMenu, new IncreaseFontSize());
        createItem(windowMenu, new DecreaseFontSize());
        createItem(windowMenu, new CascadeWindows());
        createItem(windowMenu, new CloseAllWindows());
        createItem(windowMenu, new DetachCurrentTab());
        ComponentFactory.addSubmenu(windowMenu, createLAFMenu());
        windowMenu.addSeparator();
        Project project = ProjectManager.getProjectManager().getCurrentProject();
        if (project != null) {
            Iterator i = project.getOpenWindows().iterator();
            while (i.hasNext()) {
                JFrame jframe = (JFrame) i.next();
                createItem(windowMenu, new JFrameToFront(jframe));
            }
        }
    }
    public String toString() {
        return StringUtilities.getClassName(this);
    }
}

class PluginComparator implements Comparator {
    public int compare(Object o1, Object o2) {
        Plugin p1 = (Plugin) o1;
        Plugin p2 = (Plugin) o2;
        return p1.getName().compareTo(p2.getName());
    }
}