/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2001.  All Rights Reserved.
 *
 * Protege-2000 was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu
 *
 * Contributor(s):
 */

package edu.stanford.smi.protege.util;

import java.awt.*;
import java.awt.dnd.*;
import java.awt.event.*;
import java.beans.*;
import java.io.*;
import java.net.*;
import java.util.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.text.*;

import edu.stanford.smi.protege.action.*;
import edu.stanford.smi.protege.resource.*;

/**
 * Factory class for making swing components, and their varients.  The use of this
 * class is not required for Protege widgets.  It is encouraged though.  This allows
 * for a single place to address swing bugs and look and feel issues.
 *
 * @author    Ray Fergerson <fergerson@smi.stanford.edu>
 * @author    Monica Crubezy <crubezy@smi.stanford.edu>
 */
public class ComponentFactory {
    public final static int STANDARD_BUTTON_HEIGHT = 25;
    public final static int STANDARD_BUTTON_WIDTH = 25;
    public final static int STANDARD_FIELD_HEIGHT = STANDARD_BUTTON_HEIGHT;
    public final static Dimension STANDARD_BUTTON_SIZE = new Dimension(STANDARD_BUTTON_WIDTH, STANDARD_BUTTON_HEIGHT);
    public final static int SMALL_BUTTON_HEIGHT = 16;
    public final static int SMALL_BUTTON_WIDTH = 16;
    public final static Dimension SMALL_BUTTON_SIZE = new Dimension(SMALL_BUTTON_WIDTH, SMALL_BUTTON_HEIGHT);
    private static File _lastDirectory;

    private static int _offset;
    private final static int OFFSET_SIZE = 25;

    private static class DisposableFrame extends JFrame implements Disposable {
        public DisposableFrame() {
            ComponentUtilities.registerWindow(this);
            enableEvents(AWTEvent.WINDOW_EVENT_MASK);
        }

        public void processWindowEvent(WindowEvent event) {
            if (event.getID() == WindowEvent.WINDOW_CLOSED) {
                ComponentUtilities.deregisterWindow(this);
            }
            super.processWindowEvent(event);
        }
    }

    public static void addMenuItem(JMenu menu, Action action) {
        JMenuItem item = new JMenuItem(action) {
            public String getText() {
                String text = null;
                Action action = getAction();
                if (action != null) {
                    text = (String) action.getValue(Action.NAME);
                }
                return text;
            }
        };
        menu.add(item);
        fixFont(item);
        if (item.getIcon() == null) {
            item.setIcon(Icons.getBlankIcon());
        }
        item.setDisabledIcon((Icon) action.getValue(Action.SMALL_ICON));
    }

    public static void addMenuItemNoIcon(JMenu menu, Action action) {
        // menu.add(action);
        addMenuItem(menu, action);
    }

    public static void addSubmenu(JMenu menu, JMenu submenu) {
        submenu.setIcon(Icons.getBlankIcon());
        // submenu.setHorizontalTextPosition(JMenu.RIGHT);
        menu.add(submenu);
    }

    public static JToggleButton addToggleToolBarButton(FakeToolBar toolBar, final Action action) {
        JToggleButton button = new JToggleButton((Icon) action.getValue(Action.SMALL_ICON));
        addToolBarButton(toolBar, action, button);
        return button;
    }

    public static JButton addToolBarButton(FakeToolBar toolBar, final Action action) {
        JButton button = new JButton((Icon) action.getValue(Action.SMALL_ICON));
        addToolBarButton(toolBar, action, button);
        return button;
    }

    public static void addToolBarButton(FakeToolBar toolBar, final Action action, final AbstractButton button) {
        button.addActionListener(action);
        String toolTip = (String) action.getValue(Action.SHORT_DESCRIPTION);
        if (toolTip == null) {
            toolTip = (String) action.getValue(Action.NAME);
        }
        button.setToolTipText(toolTip);
        button.setFocusPainted(false);

        action.addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent event) {
                button.setEnabled(action.isEnabled());
            }
        });
        button.setEnabled(action.isEnabled());
        setToolbarButtonBorder(button);

        toolBar.addButton(button);
    }

    public static JButton addToolBarButton(JToolBar bar, Action action) {
        JButton button = new JButton(action) {
            /* This is tricky because the JComponent.setToolTipText 
             * relies on this method for some side effects. 
             */
            public String getToolTipText() {
                String text = super.getToolTipText();
                if (text != null) {
                    text = (String) getAction().getValue(Action.NAME);
                }
                return text;
            }
        };
        addButton(bar, action, button);
        return button;
    }

    public static JToggleButton addToolBarToggleButton(JToolBar bar, Action action) {
        JToggleButton button = new JToggleButton(action);
        addButton(bar, action, button);
        return button;
    }

    private static void addButton(JToolBar bar, Action action, AbstractButton button) {
        button.setToolTipText((String) action.getValue(Action.NAME));
        button.setText("");
        button.setFocusPainted(false);

        button.setPreferredSize(STANDARD_BUTTON_SIZE);
        button.setMaximumSize(STANDARD_BUTTON_SIZE);
        button.setMinimumSize(STANDARD_BUTTON_SIZE);
        button.setSize(STANDARD_BUTTON_SIZE);

        setToolbarButtonBorder(button);

        bar.add(button);
    }

    private static MouseListener mouseListener = new MouseAdapter() {
        public void mouseEntered(MouseEvent event) {
            AbstractButton button = getButton(event);
            button.setBorderPainted(button.isEnabled());
        }
        public void mouseExited(MouseEvent event) {
            getButton(event).setBorderPainted(false);
        }

        private AbstractButton getButton(MouseEvent event) {
            return (AbstractButton) event.getSource();
        }

    };

    public static void setToolbarButtonBorder(final AbstractButton button) {
        if (!Boolean.getBoolean("buttons.paintborders")) {
            button.addMouseListener(mouseListener);
            button.setBorderPainted(false);
        }
    }

    private static void adjustPosition(Component c) {
        _offset = (_offset + 1) % 4;
        Point p = c.getLocation();
        p.x += _offset * OFFSET_SIZE;
        p.y += _offset * OFFSET_SIZE;
        c.setLocation(p);
    }

    private static Dimension buttonPreferredHeightSize(Dimension d) {
        d.height = STANDARD_BUTTON_HEIGHT;
        return d;
    }

    private static void configureList(JList list, Action action, boolean enableDragAndDrop) {
        list.setModel(new SimpleListModel());
        if (action != null) {
            list.addMouseListener(new DoubleClickActionAdapter(action));
        }
        if (enableDragAndDrop) {
            setupDragAndDrop(list);
            ComponentUtilities.setDragAndDropEnabled(list, enableDragAndDrop);
        }
        list.setCellRenderer(new DefaultRenderer());
    }

    public static void configureTree(JTree tree, Action action) {
        if (action != null) {
            tree.addMouseListener(new DoubleClickActionAdapter(action));
        }

        tree.setRootVisible(false);
        tree.setShowsRootHandles(false);

        // this causes reasonable looking parent to child lines to be drawn
        // tree.putClientProperty("JTree.lineStyle", "Angled");
    }

    public static JButton createButton(Action action) {
        JButton button = new JButton() {
            public Dimension getPreferredSize() {
                return buttonPreferredHeightSize(super.getPreferredSize());
            }
        };
        initializeAbstractButton(button, action);
        return button;
    }

    public static JPanel createButtonPreferredHeightPanel() {
        JPanel panel = new JPanel() {
            public Dimension getPreferredSize() {
                return buttonPreferredHeightSize(super.getPreferredSize());
            }
        };
        return panel;
    }

    public static JCheckBox createCheckBox() {
        return createCheckBox("");
    }

    public static JCheckBox createCheckBox(String s) {
        JCheckBox checkBox = new JCheckBox(s);
        fixFont(checkBox);
        return checkBox;
    }

    public static JComboBox createComboBox() {
        JComboBox comboBox = new JComboBox() {
            public Dimension getPreferredSize() {
                return fieldPreferredHeightSize(super.getPreferredSize());
            }
        };
        fixFont(comboBox);
        return comboBox;
    }

    public static FakeToolBar createFakeToolBar() {
        return new FakeToolBar(SwingConstants.RIGHT);
    }

    public static FakeToolBar createFakeToolBar(Dimension d) {
        return new FakeToolBar(SwingConstants.RIGHT, d);
    }

    public static JFileChooser createFileChooser(String description, String extension) {
        if (_lastDirectory == null) {
            _lastDirectory = ApplicationProperties.getApplicationDirectory();
        }

        JFileChooser chooser = new JFileChooser(_lastDirectory) {
            public int showDialog(Component c, String s) {
                int rval = super.showDialog(c, s);
                _lastDirectory = getCurrentDirectory();
                return rval;
            }
        };
        chooser.setDialogTitle(description);
        if (extension == null) {
            chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
        } else {
            String text = Text.getProgramName() + " " + description;
            chooser.setFileFilter(new ExtensionFilter(extension, text));
        }
        return chooser;
    }

    public static JFrame createMainFrame() {
        JFrame frame = new JFrame();
        initializeFrame(frame);
        return frame;
    }

    public static JFrame createFrame() {
        JFrame frame = new DisposableFrame();
        initializeFrame(frame);
        return frame;
    }

    private static void initializeFrame(JFrame frame) {
        frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        Image nerd = Icons.getNerd16x16Image();
        frame.setIconImage(nerd);
    }

    public static JLabel createLabel() {
        JLabel label = new JLabel();
        fixFont(label);
        return label;
    }

    public static JLabel createLabel(String s) {
        JLabel label = createLabel();
        label.setText(s);
        return label;
    }

    public static JLabel createLabel(Icon icon) {
        JLabel label = createLabel();
        label.setIcon(icon);
        return label;
    }

    public static JLabel createLabel(Icon icon, int alignment) {
        JLabel label = createLabel(icon);
        label.setHorizontalAlignment(alignment);
        return label;
    }

    public static JLabel createLabel(String s, int alignment) {
        JLabel label = createLabel(s);
        label.setHorizontalAlignment(alignment);
        return label;
    }

    public static LabeledComponent createLabeledScrollComponent(
        String label,
        JComponent basicComponent,
        Dimension preferredSize) {
        LabeledComponent component = new LabeledComponent(label, ComponentFactory.createScrollPane(basicComponent));
        if (preferredSize != null) {
            component.setPreferredSize(preferredSize);
        }
        return component;
    }

    public static LabeledComponent createLabeledScrollComponent(
        String label,
        JComponent basicComponent,
        Dimension preferredSize,
        JComponent headerComponent,
        Collection headerButtons,
        JComponent footerComponent) {
        LabeledComponent component = createLabeledScrollComponent(label, basicComponent, preferredSize);
        if (headerComponent != null) {
            component.setHeaderComponent(headerComponent);
        }
        if (headerButtons != null) {
            Iterator buttonList = headerButtons.iterator();
            Action button;
            while (buttonList.hasNext()) {
                if (buttonList != null) {
                    button = (Action) buttonList.next();
                    component.addHeaderButton(button);
                }
            }
        }
        if (footerComponent != null) {
            component.setFooterComponent(footerComponent);
        }
        return component;
    }

    public static JSplitPane createLeftRightSplitPane() {
        return createLeftRightSplitPane(true);
    }

    public static JSplitPane createLeftRightSplitPane(boolean autoResize) {
        return createSplitPane(JSplitPane.HORIZONTAL_SPLIT, autoResize);
    }

    public static JSplitPane createTopBottomSplitPane(JComponent top, JComponent bottom) {
        JSplitPane splitPane = createTopBottomSplitPane(true);
        splitPane.setTopComponent(top);
        splitPane.setBottomComponent(bottom);
        return splitPane;
    }
    public static JSplitPane createLeftRightSplitPane(JComponent left, JComponent right) {
        JSplitPane splitPane = createLeftRightSplitPane(true);
        splitPane.setLeftComponent(left);
        splitPane.setRightComponent(right);
        return splitPane;
    }
    public static JList createList(Action action) {
        return createList(action, false);
    }

    public static JList createList(Action action, boolean enableDragAndDrop) {
        return createSelectableList(action, enableDragAndDrop);
    }

    public static JMenuItem createMenuItem(String s) {
        JMenuItem item = new JMenuItem(s);
        fixFont(item);
        return item;
    }

    public static JMenu createMenu() {
        JMenu menu = new JMenu();
        fixFont(menu);
        return menu;
    }

    public static JMenu createMenu(String text) {
        JMenu menu = createMenu();
        menu.setText(text);
        return menu;
    }

    public static JPanel createPanel() {
        JPanel panel = new JPanel();
        return panel;
    }

    public static JPasswordField createPasswordField() {
        JPasswordField passwordField = new JPasswordField() {
            public Dimension getPreferredSize() {
                return fieldPreferredHeightSize(super.getPreferredSize());
            }
        };
        return passwordField;
    }

    public static JRadioButtonMenuItem createRadioButtonMenuItem(Action action) {
        JRadioButtonMenuItem item = new JRadioButtonMenuItem();
        fixFont(item);
        initializeAbstractButton(item, action);
        return item;
    }

    public static JScrollPane createScrollPane(JComponent c) {
        JScrollPane pane = new JScrollPane(c);
        return pane;
    }

    public static JScrollPane createScrollPane(final JTable table) {
        JScrollPane pane = new JScrollPane(table) {
            public void updateUI() {
                super.updateUI();
                JViewport vp = getViewport();
                // workaround bug where the table is not initialized properly
                Color bgcolor = (table != null) ? table.getBackground() : Color.WHITE;
                vp.setBackground(bgcolor);
            }
        };
        return pane;
    }

    public static SelectableList createSelectableList(Action action) {
        return createSelectableList(action, false);
    }

    public static SelectableList createSelectableList(Action action, boolean enableDragAndDrop) {
        SelectableList list = new SelectableList();
        configureList(list, action, enableDragAndDrop);
        return list;
    }

    public static SelectableTable createSelectableTable(Action action) {
        SelectableTable table = new SelectableTable();
        if (action != null) {
            table.addMouseListener(new DoubleClickActionAdapter(action));
        }
        table.setShowGrid(false);
        table.setIntercellSpacing(new Dimension(0, 0));
        table.setColumnSelectionAllowed(false);
        table.setAutoResizeMode(JTable.AUTO_RESIZE_LAST_COLUMN);
        table.setAutoCreateColumnsFromModel(false);
        table.setDefaultEditor(Object.class, null);
        return table;
    }

    public static SelectableTree createSelectableTree(Action action) {
        return createSelectableTree(action, null);
    }

    public static SelectableTree createSelectableTree(Action action, LazyTreeRoot root) {
        SelectableTree tree = new SelectableTree(action, root);
        return tree;
    }

    public static SelectableList createSingleItemList(Action action) {
        SelectableList list = (SelectableList) createList(action);
        list.setPreferredSize(new Dimension(1, 24));
        list.setFixedCellHeight(20);
        list.setBorder(BorderFactory.createEtchedBorder());
        return list;
    }

    private static JSplitPane createSplitPane(int direction, final boolean autoResize) {
        JSplitPane pane = new JSplitPane(direction, true) {
            public void addImpl(Component c, Object o, int i) {
                if (c instanceof JComponent) {
                    ((JComponent) c).setMinimumSize(new Dimension(50, 50));
                }
                super.addImpl(c, o, i);
            }
            public void updateUI() {
                super.updateUI();
                setBorder(null);
            }
        };
        pane.setOneTouchExpandable(true);
        return pane;
    }

    public static Border createStandardBorder() {
        return BorderFactory.createEmptyBorder(5, 5, 5, 5);
    }

    public static JTabbedPane createTabbedPane(final boolean addBorder) {
        JTabbedPane pane = new JTabbedPane() {
            public void addImpl(Component component, Object constraints, int index) {
                if (addBorder) {
                    JComponent c = (JComponent) component;
                    c.setBorder(BorderFactory.createCompoundBorder(createStandardBorder(), c.getBorder()));
                }
                super.addImpl(component, constraints, index);
            }

        };
        fixFont(pane);
        return pane;
    }

    private static void fixFont(Component c) {
        // c.setFont(c.getFont().deriveFont(Font.PLAIN));
    }

    public static JTable createTable(Action action) {
        return createSelectableTable(action);
    }

    public static JTextArea createTextArea() {
        JTextArea area = new JTextArea() {
            public void setText(String text) {
                super.setText(text);
                setCaretPosition(0);
                repaint();
            }
        };
        area.setLineWrap(true);
        area.setWrapStyleWord(true);
        configureTextComponent(area);
        return area;
    }

    public static JTextField createTextField() {
        JTextField textField = new JTextField() {
            public Dimension getPreferredSize() {
                return fieldPreferredHeightSize(super.getPreferredSize());
            }
        };
        configureTextComponent(textField);
        return textField;
    }

    private static void configureTextComponent(JTextComponent component) {
        Keymap keymap = component.getKeymap();
        KeyStroke stroke = KeyStroke.getKeyStroke(KeyEvent.VK_X, InputEvent.ALT_MASK);
        Action action = new ConvertUnicodeSequenceAction();
        keymap.addActionForKeyStroke(stroke, action);
    }

    public static JTextPane createTextPane() {
        JTextPane pane = new JTextPane();
        return pane;
    }

    public static JToggleButton createToggleButton(Action action) {
        JToggleButton button = new JToggleButton() {
            public Dimension getPreferredSize() {
                return buttonPreferredHeightSize(super.getPreferredSize());
            }
        };
        initializeAbstractButton(button, action);
        return button;
    }

    public static JToolBar createToolBar() {
        JToolBar bar = new JToolBar();
        bar.setFloatable(false);
        return bar;
    }

    public static JSplitPane createTopBottomSplitPane() {
        return createTopBottomSplitPane(true);
    }

    public static JSplitPane createTopBottomSplitPane(boolean autoResize) {
        return createSplitPane(JSplitPane.VERTICAL_SPLIT, autoResize);
    }

    public static JTree createTree(Action action) {
        return createSelectableTree(action);
    }

    public static JWindow createWindow() {
        JWindow window = new JWindow();
        return window;
    }

    private static Dimension fieldPreferredHeightSize(Dimension d) {
        d.height = STANDARD_FIELD_HEIGHT;
        return d;
    }

    public static JComponent getCloseButtonPanel(final JFrame frame) {
        JComponent c = new JPanel();
        c.setLayout(new FlowLayout());
        JButton button = createButton(new AbstractAction("Close", Icons.getCloseIcon()) {
            public void actionPerformed(ActionEvent event) {
                ComponentUtilities.closeWindow(frame);
            }
        });
        c.add(button);
        return c;
    }

    private static void initializeAbstractButton(final AbstractButton button, final Action action) {
        action.addPropertyChangeListener(new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent event) {
                button.setEnabled(action.isEnabled());
            }
        });
        button.setEnabled(action.isEnabled());
        button.addActionListener(action);
        button.setIcon((Icon) action.getValue(Action.SMALL_ICON));
        button.setAlignmentX(0.5f);
        button.setAlignmentY(0.5f);
        button.setText((String) action.getValue(Action.NAME));
        button.setToolTipText((String) action.getValue(Action.SHORT_DESCRIPTION));
        button.setHorizontalTextPosition(AbstractButton.RIGHT);
        fixFont(button);
    }

    public static void removeToolBarButton(FakeToolBar toolBar, JButton button) {
        int nStartComponents = toolBar.getComponentCount();
        toolBar.remove(button);
        int nEndComponents = toolBar.getComponentCount();
        if (nEndComponents != nStartComponents) {
            toolBar.setLayout(new GridLayout(1, nEndComponents));
        }
    }

    private static void setupDragAndDrop(JList list) {

        DragSource.getDefaultDragSource().createDefaultDragGestureRecognizer(
            list,
            DnDConstants.ACTION_COPY_OR_MOVE,
            new DefaultListDragSourceListener());
        new DropTarget(list, DnDConstants.ACTION_COPY_OR_MOVE, new ListTarget());
    }

    public static JFrame showInFrame(Component panel, String title) {
        JFrame frame = createFrame();
        frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        Container c = frame.getContentPane();
        c.setLayout(new BorderLayout());
        c.add(panel, BorderLayout.CENTER);
        c.add(getCloseButtonPanel(frame), BorderLayout.SOUTH);
        frame.pack();
        frame.setTitle(title);
        ComponentUtilities.center(frame);
        adjustPosition(frame);
        frame.setVisible(true);
        return frame;
    }

    public static JEditorPane createEditorPane() {
        return new JEditorPane();
    }

    public static JEditorPane createHTMLBrowser(URL url) {
        JEditorPane pane = new JEditorPane();
        pane.setEditable(false);
        pane.addHyperlinkListener(new HyperlinkListener() {
            public void hyperlinkUpdate(HyperlinkEvent e) {
                if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                    SystemUtilities.showHTML(e.getURL().toString());
                }
            }
        });
        if (url != null) {
            setPage(pane, url);
        }
        return pane;
    }

    public static void setPage(JEditorPane pane, URL url) {
        try {
            pane.setPage(url);
        } catch (IOException e) {
            Log.exception(e, ComponentFactory.class, "setPage", pane, url);
        }
    }

}